#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "gdal_priv.h"
#include "gdal_alg.h"
#include <iostream>
#include <fstream>
#include <stdlib.h>

using namespace std;


#define CHECK(call)\
{\
  const cudaError_t error=call;\
  if(error!=cudaSuccess)\
  {\
      printf("ERROR: %s:%d,",__FILE__,__LINE__);\
      printf("code:%d,reason:%s\n",error,cudaGetErrorString(error));\
      exit(1);\
  }\
}

#define TILE_SIZE 3601
#define map_bsize(dtype) (TILE_SIZE * TILE_SIZE * sizeof(dtype))
#define cufree(buf) if(buf) {cudaFree(buf); buf=0;}
#define sfree(buf) if(buf) {free(buf); buf=0;}

struct _tile_info
{
	int ndim;  //γ
	int nsubgrid;
	double topleft[2]; //Ͼγ
	double btnright[2];//¾γ
	double cellSize, cellSizeT;//ԪСԪ 
	double lonext, latext; //γȷΧ3*3㣩ķΧ
	double loncen, latcen; //3*3ؿĵ㾭γ
	double* tilequads;
	double dheoffs[5];   //ĸ߳½ֵıdh(u,v)ƫ
	double transform[6]; //transϢ
	short* demtiles33[9]; //߳

};

#define M_PIPI   3.1415926535
#define M_2PI       6.28318530717958647692528676655900576839433L
#define M_SQRT3     1.7320508075688772935274463415058723669428L
#define M_ONE       1.00000000000000000000000000000000000000000L
#define M_ZERO      0.00000000000000000000000000000000000000000L
#define M_HALF      0.50000000000000000000000000000000000000000L
#define M_SQRT3_2   0.8660254037844386467637231707529361834714L
#define M_SIN60     0.8660254037844386467637231707529361834714L
#define M_PI_180    0.0174532925199432957692369076848861271111L
#define M_180_PI    57.29577951308232087679815481410517033240547L
#define PRECISION 0.0000000000005L 
#define originXOff 0.6022955029L
#define originYOff 0.3477354707L
#define icosaEdge 1.2045910058L
#define GH 0.6283185307179586231995926937088370323181L
#define R1 0.9103832815L
#define R1S 0.8287977192347083521539730099902953952551L
#define DH 0.6523581397730289221925659148837439715862L
#define cot30 1.7320508075688774152212090484681539237499L
#define tanDH 0.7639320224822535587705374382494483143091L
#define cosDH  0.7946544722986497166772323907935060560703
#define sinGH  0.5877852522924731371034567928290925920010
#define cosGH  0.8090169943749474512628694355953484773636
//ϵп
#define rowStep 2048
#define colStep 2048
//HexTiffز
#define offsetXMax 4095
#define offsetYMax 4095

#define xMin -2048
#define yMin -2048
//ֵ
short* values = new short[offsetXMax * offsetYMax];
//Ƶ
//Ƶ
#define ControlPointsNum 9
double* controlPointLon = new double[ControlPointsNum];
double* controlPointLat = new double[ControlPointsNum];
double* controlPointRow = new double[ControlPointsNum];
double* controlPointCol = new double[ControlPointsNum];

//ϵֱϵתǶ
#define RowAngle -60
#define ColAngle 60

#define toler 0.0000000005L
const long double tol = 0.000000000000001L;
const int level = 18;
const int tileNum = pow(2, level - 11);

//ļȡ
char dir_read[60] = "/home/ougeng/lmdoc/cuda_pdop/data";//aster洢·
const char* str_namefmts[4] = { "ASTGTM2_N%02dE%03d", "ASTGTM2_N%02dW%03d", "ASTGTM2_S%02dE%03d", "ASTGTM2_S%02dW%03d" };
const char* str_postfmt = "_dem.tif";

char dir_save[128] = "result";//ļ·
char save_format[5] = ".bin";

const char* pszFormat = "GTiff";
GDALDriver* poDriver;

//3*3ؿϢ
_tile_info  tileinfo;
_tile_info* tileinfo_d;

//γCPU
const int gridNum = pow(2, 11 * 2);
const int gridNumSize = gridNum * sizeof(double);
double* lonArray = 0;
double* latArray = 0;
double* elev_h = 0;
double* lonArray_d = 0;
double* latAttay_d = 0;
double* elev_d = 0;

void* h_elvtiles33[9] = { 0,0,0, 0,0,0, 0,0,0 };
int  imsktile33[9] = { 0,0,0,0,0,0,0,0,0 };
char projdesc[512];
int inew_cols = 0;


//ο α ڽ keep Xƽ Yƽ תϵ
double vertTable[504] = {

	// vert 0
	2,  1, 0, 1,  -M_HALF, -M_SIN60,  3,
	1,  0, 1, 1,   -M_ONE,   M_ZERO,  2,
	5,  4, 2, 1,  -M_HALF,  M_SIN60,  1,
	-1, -1, 3, 0, -M_HALF,  M_SIN60,  1,
	4,  3, 4, 1,  M_ONE,   M_ZERO, -1,
	3,  2, 5, 1,  M_HALF, -M_SIN60, -2,

	// vert 1
	  1,  0, 0, 1,  M_ZERO,  M_ZERO,  1,
	  1,  5, 1, 1,  -M_HALF, -M_SIN60, 4,
	  10, 14, 2, 1,  -M_HALF,  M_SIN60, 1,
	  -1, -1, 3, 0, -M_HALF,  M_SIN60, 1,
	  5,  9, 4, 1,  M_ZERO,  M_ZERO,  3,
	  5,  4, 5, 1,  M_ONE,   M_ZERO,  0,

	  // vert 2
	   2,  1, 0, 1,  M_ZERO,  M_ZERO,  1,
	   2,  6, 1, 1,  -M_HALF, -M_SIN60, 4,
	   10, 10, 2, 1,  -M_HALF,  M_SIN60, 1,
	   -1, -1, 3, 0, -M_HALF,  M_SIN60, 1,
	   1,  5, 4, 1,  M_ZERO,  M_ZERO,  3,
	   1,  0, 5, 1,  M_ONE,   M_ZERO,  0,

	   // vert 3

		3,  2, 0, 1,  M_ZERO,  M_ZERO,  1,
		3,  7, 1, 1,  -M_HALF, -M_SIN60, 4,
		7, 11, 2, 1,  -M_HALF,  M_SIN60, 1,
		-1, -1, 3, 0, -M_HALF,  M_SIN60, 1,
		2,  6, 4, 1,  M_ZERO,  M_ZERO,  3,
		2,  1, 5, 1,  M_ONE,   M_ZERO,  0,

		// vert 4

		 4,  3, 0, 1,  M_ZERO,  M_ZERO,  1,
		 4,  8, 1, 1,  -M_HALF, -M_SIN60, 4,
		 8, 12, 2, 1,  -M_HALF,  M_SIN60, 1,
		 -1, -1, 3, 0, -M_HALF,  M_SIN60, 1,
		 3,  7, 4, 1,  M_ZERO,  M_ZERO,  3,
		 3,  2, 5, 1,  M_ONE,   M_ZERO,  0,

		 // vert 5

		  5,  4, 0, 1,  M_ZERO,  M_ZERO,  1,
		  5,  9, 1, 1,  -M_HALF, -M_SIN60, 4,
		  9, 13, 2, 1,  -M_HALF,  M_SIN60, 1,
		  -1, -1, 3, 0, -M_HALF,  M_SIN60, 1,
		  4,  8, 4, 1,  M_ZERO,  M_ZERO,  3,
		  4,  3, 5, 1,  M_ONE,   M_ZERO,  0,

		  // vert 6

		   6, 10, 0, 1,  M_ZERO,  M_ZERO,   1,
		   6, 15, 1, 1,  -M_HALF, -M_SIN60,  4,
		   10, 19, 2, 1,  M_ZERO,  M_ZERO,  -1,
		   10, 14, 3, 1,  -M_HALF,  M_SIN60,  2,
		   -1, -1, 4, 0, -M_HALF,  M_SIN60,  1,
		   1,  5, 5, 1,  M_HALF, -M_SIN60,  4,

		   // vert 7

			7, 11, 0, 1,  M_ZERO,  M_ZERO,   1,
			7, 16, 1, 1,  -M_HALF, -M_SIN60,  4,
			6, 15, 2, 1,  M_ZERO,  M_ZERO,  -1,
			6, 10, 3, 1,  -M_HALF,  M_SIN60,  2,
			-1, -1, 4, 0, -M_HALF,  M_SIN60,  1,
			2,  6, 5, 1,  M_HALF, -M_SIN60,  4,

			// vert 8

			 8, 12, 0, 1,  M_ZERO,  M_ZERO,   1,
			 8, 17, 1, 1,  -M_HALF, -M_SIN60,  4,
			 7, 16, 2, 1,  M_ZERO,  M_ZERO,  -1,
			 7, 11, 3, 1,  -M_HALF,  M_SIN60,  2,
			 -1, -1, 4, 0, -M_HALF,  M_SIN60,  1,
			 3,  7, 5, 1,  M_HALF, -M_SIN60,  4,

			 // vert 9

			  9, 13, 0, 1,  M_ZERO,  M_ZERO,   1,
			  9, 18, 1, 1, -M_HALF, -M_SIN60,  4,
			  8, 17, 2, 1,  M_ZERO,  M_ZERO,  -1,
			  8, 12, 3, 1,  -M_HALF,  M_SIN60,  2,
			  -1, -1, 4, 0, -M_HALF,  M_SIN60,  1,
			  4,  8, 5, 1,  M_HALF, -M_SIN60,  4,

			  // vert 10

				10, 14, 0, 1,  M_ZERO,  M_ZERO,   1,
				10, 19, 1, 1,  -M_HALF, -M_SIN60,  4,
				9, 18, 2, 1,  M_ZERO,  M_ZERO,  -1,
				9, 13, 3, 1,  -M_HALF,  M_SIN60,  2,
				-1, -1, 4, 0, -M_HALF,  M_SIN60,  1,
				5,  9, 5, 1,  M_HALF, -M_SIN60,  4,

				// vert 11 

				 8, 17, 0, 1,  -M_HALF, -M_SIN60,  3,
				 9, 18, 1, 1,  -M_ONE,   M_ZERO,   2,
				 10, 19, 2, 1,  -M_HALF,  M_SIN60,  1,
				 6, 15, 3, 1,  M_HALF,  M_SIN60,  0,
				 -1, -1, 4, 0, M_ZERO,  M_ZERO,   0,
				 7, 16, 5, 1,  M_HALF, -M_SIN60, -2
};

//ʮβ
//lon lat λ תϵ ƽX ƽY οID
double triParaments[140] = {
-2.513274,0.918438,0.000000,1,0.000000,0.000000,1,
-1.256637,0.918438,0.000000,1,0.000000,0.000000,2,
-0.000000,0.918438,0.000000,1,0.000000,0.000000,3,
1.256637,0.918438,-0.000000,1,0.000000,0.000000,4,
2.513274,0.918438,-0.000000,1,0.000000,0.000000,5,
-2.513274,0.188711,3.141593,4,-0.500000,-0.866025,1,
-1.256637,0.188711,3.141593,4,-0.500000,-0.866025,2,
-0.000000,0.188711,3.141593,4,-0.500000,-0.866025,3,
1.256637,0.188711,3.141593,4,-0.500000,-0.866025,4,
2.513274,0.188711,3.141593,4,-0.500000,-0.866025,5,
-1.884956,-0.188711,-0.000000,1,0.000000,0.000000,6,
-0.628319,-0.188711,-0.000000,1,0.000000,0.000000,7,
0.628319,-0.188711,0.000000,1,0.000000,0.000000,8,
1.884956,-0.188711,-0.000000,1,0.000000,0.000000,9,
3.141593,-0.188711,-0.000000,1,0.000000,0.000000,10,
-1.884956,-0.918438,-3.141593,4,-0.500000,-0.866025,6,
-0.628319,-0.918438,-3.141593,4,-0.500000,-0.866025,7,
0.628319,-0.918438,3.141593,4,-0.500000,-0.866025,8,
1.884956,-0.918438,3.141593,4,-0.500000,-0.866025,9,
-3.141593,-0.918438,-3.141593,4,-0.500000,-0.866025,10
};

//GPU˲Ϣ
const int triSize = 140 * sizeof(double);
const int vertTriSize = 12 * 42 * sizeof(double);

double* vertTable_d = 0;
double* triParaments_d = 0;


/*************************************************************/
int  load_one_tile(int ilon, int ilat, int ishp, void* buffer, double* ptransform);
int  lonlat_ishp(int& ilon, int& ilat);
int  load_tile33_all(int ilon, int ilat);
void loadGridData(int baseQuad, int rowIndex, int colIndex, double* lon, double* lat);
void bufferInit();
void bufferDelete();
int map_buffer(void* dst, void* src, int bsize, int itodevice);
int upd_demtile33_all();
int cudop_upd_tileinfo(_tile_info* pinfo);
__device__  double get_elev_near_min(double lon, double lat, _tile_info* tinfo_d);
__global__ void dem2Hex(double* d_lonArray, double* d_latArray, double* d_elev, _tile_info* tinfo_d);
/*************************************************************/

/***************************************************************************************/
/* code2Geo */
//кת
void invQuantify_H(int i, int j, double* add)
{
	add[0] = i - 0.5 * j;
	add[1] = j * 1.5 * M_SQRT3 / 3;
}
//
int computeSubTriangle_H(const double& x, const double& y)
{

	long double xs = M_SQRT3 * x;

	long double xpp = xs + tol,
		xmp = -xs + tol,
		xpm = xs - tol,
		xmm = -xs - tol;

	if (y >= xmm && y > xpp)
	{
		return 0;
	}

	if ((fabs(y) <= tol && fabs(x) <= tol) ||
		(y <= xpp && y >= (0.0 - tol)))
	{
		return 1;
	}

	if (y < (0.0 - tol) && y > xmp)
	{
		return 2;
	}

	if (y <= xmp && y < xpm)
	{
		return 3;
	}

	if (y >= xpm && y < (0.0 - tol))
	{
		return 4;
	}

	if (y >= (0.0 - tol) && y < xmm)
	{
		return 5;
	}

	return -1;
}
//ʩ͵·
void snyderInv_H(double* xyCoord, int triangle, double* lonlatCoord)
{
	double ddazh, ph, fh, azh, azh1, dazh, h, fazh, flazh, dz, z;
	double sinlat, sinlon;
	double azh0;

	//cent
	double tri_lon = triParaments[triangle * 7 + 0];
	double tri_lat = triParaments[triangle * 7 + 1];

	//pt
	double ptX = xyCoord[0] * icosaEdge - originXOff;
	double ptY = xyCoord[1] * icosaEdge - originYOff;

	ddazh = triParaments[triangle * 7 + 2];

	if ((fabs(ptX) < PRECISION) && (fabs(ptY) < PRECISION))
	{
		lonlatCoord[0] = tri_lon;
		lonlatCoord[1] = tri_lat;
	}
	else
	{
		ph = sqrt(ptX * ptX + ptY * ptY);
		azh1 = atan2(ptX, ptY);

		if (azh1 < 0.0L)
			azh1 = azh1 + 2 * M_PIPI;
		azh0 = azh1;
		if ((azh1 > 120.0L * M_PI_180) && (azh1 <= 240.0L * M_PI_180))
			azh1 = azh1 - 120.0L * M_PI_180;
		if (azh1 > 240.0L * M_PI_180)
			azh1 = azh1 - 240.0L * M_PI_180;
		azh = azh1;

		if (fabs(azh1) > PRECISION)
		{
			long double agh = R1S * tanDH * tanDH / (2.0L * (1.0L / tan(azh1) + cot30));
			dazh = 1.0;
			while (fabs(dazh) > PRECISION)
			{
				h = acos(sin(azh) * sinGH * cosDH - cos(azh) * cosGH);
				fazh = agh - azh - GH - h + M_PIPI;
				flazh = ((cos(azh) * sinGH * cosDH + sin(azh) * cosGH) / sin(h)) - 1.0;
				dazh = -fazh / flazh;
				azh = azh + dazh;
			}
		}
		else azh = azh1 = 0.0;

		dz = atan2f(tanDH, cos(azh) + cot30 * sin(azh));
		fh = tanDH / (2.0L * (cos(azh1) + cot30 * sin(azh1)) * sinf(dz / 2.0L));
		z = 2.0 * asinf(ph / (2.0L * R1 * fh));
		if ((azh0 >= 120 * M_PI_180) && (azh0 < 240.0L * M_PI_180)) azh = azh + 120 * M_PI_180;
		if (azh0 >= 240.0L * M_PI_180) azh = azh + 240.0L * M_PI_180;

		azh += ddazh;

		while (azh <= -M_PIPI) azh += M_2PI;
		while (azh > M_PIPI) azh -= M_2PI;

		sinlat = sin(tri_lat) * cos(z) + cos(tri_lat) * sin(z) * cos(azh);
		if (sinlat > M_ONE) sinlat = M_ONE;
		if (sinlat < -M_ONE) sinlat = -M_ONE;
		lonlatCoord[1] = asin(sinlat);

		if (fabs(fabs(lonlatCoord[1]) - 1.57079632675) < PRECISION)
		{
			lonlatCoord[1] = (lonlatCoord[1] > M_ZERO) ? 1.57079632675 : -1.57079632675;
			lonlatCoord[0] = M_ZERO;
		}
		else
		{
			sinlon = sin(azh) * sin(z) / cos(lonlatCoord[1]);
			long double coslon = (cos(z) - sin(tri_lat) * sin(lonlatCoord[1])) /
				cos(tri_lat) / cos(lonlatCoord[1]);
			if (sinlon > M_ONE) sinlon = M_ONE;
			if (sinlon < -M_ONE) sinlon = -M_ONE;
			if (coslon > M_ONE) coslon = M_ONE;
			if (coslon < -M_ONE) coslon = -M_ONE;
			lonlatCoord[0] = tri_lon + asinf(sinlon);
			lonlatCoord[0] = tri_lon + atan2f(sinlon, coslon);
			if (lonlatCoord[0] <= -M_PIPI) lonlatCoord[0] += M_2PI;
			if (lonlatCoord[0] >= M_PIPI) lonlatCoord[0] -= M_2PI;
		}
	}
}
//תγ
void convertTypeAddress_H(double* cent, double scaleFac, double* trans, double rotDeg, int quadNum, double* result)
{
	double tempRes[2] = { 0.0,0.0 };

	double x_temp = cent[0];
	double y_temp = cent[1];
	x_temp /= scaleFac;
	y_temp /= scaleFac;

	tempRes[0] = x_temp;
	tempRes[1] = y_temp;

	result[0] = x_temp;
	result[1] = y_temp;

	int subTri = computeSubTriangle_H(x_temp, y_temp);

	if (vertTable[quadNum * 42 + subTri * 7 + 3])
	{
		double rotateAng = vertTable[quadNum * 42 + subTri * 7 + 6] * -60.0;
		while (rotateAng < M_ZERO) rotateAng += 360.0L;
		while (rotateAng >= 360.0L) rotateAng -= 360.0L;
		rotateAng *= M_PI_180;

		x_temp += vertTable[quadNum * 42 + subTri * 7 + 4];
		y_temp += vertTable[quadNum * 42 + subTri * 7 + 5];

		tempRes[0] = x_temp * cos(rotateAng) - y_temp * sin(rotateAng);
		tempRes[1] = x_temp * sin(rotateAng) + y_temp * cos(rotateAng);

		int triangle = vertTable[quadNum * 42 + subTri * 7 + 1];

		double lonlatCoord[2] = { 0.0,0.0 };

		//һ㲻
		snyderInv_H(tempRes, triangle, lonlatCoord);

		if (fabsf(lonlatCoord[1] * M_180_PI) >= (90.0 - PRECISION))
		{
			lonlatCoord[0] = 0.0;
		}
		else
		{
			while (lonlatCoord[0] * M_180_PI <= -180.0)
			{
				lonlatCoord[0] = (lonlatCoord[0] * M_180_PI + 360.0) * M_PI_180;
			}
			while (lonlatCoord[0] * M_180_PI > 180.0)
			{
				lonlatCoord[0] = (lonlatCoord[0] * M_180_PI - 360.0) * M_PI_180;
			}
		}

		result[0] = lonlatCoord[0];
		result[1] = lonlatCoord[1];

	}
}

void getCenter_H(int quadNum, int i, int j, int level, double* result)
{
	double cent[2] = { 0.0,0.0 };
	invQuantify_H(i, j, cent);

	//long double r = M_SQRT3 / 3;
	//long double r2 = r / 2;
	double trans[2] = { 0.0,0.0 };
	double rotDeg = 0.0;
	double scaleFac = powf(2, level);

	convertTypeAddress_H(cent, scaleFac, trans, rotDeg, quadNum, result);
}

void code2Geo_H(int basePartition, int row, int col, int level, double& lon, double& lat)
{
	double result[2] = { 0.0,0.0 };
	getCenter_H(basePartition, row, col, level, result);

	lon = result[0] * M_180_PI;
	lat = result[1] * M_180_PI;
}

/***************************************************************************************/
//кת
__device__ void invQuantify(int i, int j, double* add)
{
	add[0] = i - 0.5 * j;
	add[1] = j * 1.5 * M_SQRT3 / 3;
}
//
__device__ int computeSubTriangle(const double& x, const double& y)
{
	const long double tol = 0.000000000000001L;

	long double xs = M_SQRT3 * x;

	long double xpp = xs + tol,
		xmp = -xs + tol,
		xpm = xs - tol,
		xmm = -xs - tol;

	if (y >= xmm && y > xpp)
	{
		return 0;
	}

	if ((fabs(y) <= tol && fabs(x) <= tol) ||
		(y <= xpp && y >= (0.0 - tol)))
	{
		return 1;
	}

	if (y < (0.0 - tol) && y > xmp)
	{
		return 2;
	}

	if (y <= xmp && y < xpm)
	{
		return 3;
	}

	if (y >= xpm && y < (0.0 - tol))
	{
		return 4;
	}

	if (y >= (0.0 - tol) && y < xmm)
	{
		return 5;
	}

	return -1;
}
//ʩ͵·
__device__ void snyderInv(double* xyCoord, int triangle, double* lonlatCoord, double* triParaments_d)
{
	double ddazh, ph, fh, azh, azh1, dazh, h, fazh, flazh, dz, z;
	double sinlat, sinlon;
	double azh0;

	//cent
	double tri_lon = triParaments_d[triangle * 7 + 0];
	double tri_lat = triParaments_d[triangle * 7 + 1];

	//pt
	double ptX = xyCoord[0] * icosaEdge - originXOff;
	double ptY = xyCoord[1] * icosaEdge - originYOff;

	ddazh = triParaments_d[triangle * 7 + 2];

	if ((fabs(ptX) < PRECISION) && (fabs(ptY) < PRECISION))
	{
		lonlatCoord[0] = tri_lon;
		lonlatCoord[1] = tri_lat;
	}
	else
	{
		ph = sqrt(ptX * ptX + ptY * ptY);
		azh1 = atan2(ptX, ptY);

		if (azh1 < 0.0L)
			azh1 = azh1 + 2 * M_PIPI;
		azh0 = azh1;
		if ((azh1 > 120.0L * M_PI_180) && (azh1 <= 240.0L * M_PI_180))
			azh1 = azh1 - 120.0L * M_PI_180;
		if (azh1 > 240.0L * M_PI_180)
			azh1 = azh1 - 240.0L * M_PI_180;
		azh = azh1;

		if (fabs(azh1) > PRECISION)
		{
			long double agh = R1S * tanDH * tanDH / (2.0L * (1.0L / tan(azh1) + cot30));
			dazh = 1.0;
			while (fabs(dazh) > PRECISION)
			{
				h = acos(sin(azh) * sinGH * cosDH - cos(azh) * cosGH);
				fazh = agh - azh - GH - h + M_PIPI;
				flazh = ((cos(azh) * sinGH * cosDH + sin(azh) * cosGH) / sin(h)) - 1.0;
				dazh = -fazh / flazh;
				azh = azh + dazh;
			}
		}
		else azh = azh1 = 0.0;

		dz = atan2f(tanDH, cos(azh) + cot30 * sin(azh));
		fh = tanDH / (2.0L * (cos(azh1) + cot30 * sin(azh1)) * sinf(dz / 2.0L));
		z = 2.0 * asinf(ph / (2.0L * R1 * fh));
		if ((azh0 >= 120 * M_PI_180) && (azh0 < 240.0L * M_PI_180)) azh = azh + 120 * M_PI_180;
		if (azh0 >= 240.0L * M_PI_180) azh = azh + 240.0L * M_PI_180;

		azh += ddazh;

		while (azh <= -M_PIPI) azh += M_2PI;
		while (azh > M_PIPI) azh -= M_2PI;

		sinlat = sin(tri_lat) * cos(z) + cos(tri_lat) * sin(z) * cos(azh);
		if (sinlat > M_ONE) sinlat = M_ONE;
		if (sinlat < -M_ONE) sinlat = -M_ONE;
		lonlatCoord[1] = asin(sinlat);

		if (fabs(fabs(lonlatCoord[1]) - 1.57079632675) < PRECISION)
		{
			lonlatCoord[1] = (lonlatCoord[1] > M_ZERO) ? 1.57079632675 : -1.57079632675;
			lonlatCoord[0] = M_ZERO;
		}
		else
		{
			sinlon = sin(azh) * sin(z) / cos(lonlatCoord[1]);
			long double coslon = (cos(z) - sin(tri_lat) * sin(lonlatCoord[1])) /
				cos(tri_lat) / cos(lonlatCoord[1]);
			if (sinlon > M_ONE) sinlon = M_ONE;
			if (sinlon < -M_ONE) sinlon = -M_ONE;
			if (coslon > M_ONE) coslon = M_ONE;
			if (coslon < -M_ONE) coslon = -M_ONE;
			lonlatCoord[0] = tri_lon + asinf(sinlon);
			lonlatCoord[0] = tri_lon + atan2f(sinlon, coslon);
			if (lonlatCoord[0] <= -M_PIPI) lonlatCoord[0] += M_2PI;
			if (lonlatCoord[0] >= M_PIPI) lonlatCoord[0] -= M_2PI;
		}
	}
}
//תγ
__device__ void convertTypeAddress(double* cent, double scaleFac, double* trans, double rotDeg, int quadNum, double* result, double* triParaments_d, double* vertTable_d)
{
	double tempRes[2] = { 0.0,0.0 };

	double x_temp = cent[0];
	double y_temp = cent[1];
	x_temp /= scaleFac;
	y_temp /= scaleFac;

	tempRes[0] = x_temp;
	tempRes[1] = y_temp;

	result[0] = x_temp;
	result[1] = y_temp;

	int subTri = computeSubTriangle(x_temp, y_temp);

	if (vertTable_d[quadNum * 42 + subTri * 7 + 3])
	{
		double rotateAng = vertTable_d[quadNum * 42 + subTri * 7 + 6] * -60.0;
		while (rotateAng < M_ZERO) rotateAng += 360.0L;
		while (rotateAng >= 360.0L) rotateAng -= 360.0L;
		rotateAng *= M_PI_180;

		x_temp += vertTable_d[quadNum * 42 + subTri * 7 + 4];
		y_temp += vertTable_d[quadNum * 42 + subTri * 7 + 5];

		tempRes[0] = x_temp * cos(rotateAng) - y_temp * sin(rotateAng);
		tempRes[1] = x_temp * sin(rotateAng) + y_temp * cos(rotateAng);

		int triangle = vertTable_d[quadNum * 42 + subTri * 7 + 1];

		double lonlatCoord[2] = { 0.0,0.0 };

		//һ㲻
		snyderInv(tempRes, triangle, lonlatCoord, triParaments_d);

		if (fabsf(lonlatCoord[1] * M_180_PI) >= (90.0 - PRECISION))
		{
			lonlatCoord[0] = 0.0;
		}
		else
		{
			while (lonlatCoord[0] * M_180_PI <= -180.0)
			{
				lonlatCoord[0] = (lonlatCoord[0] * M_180_PI + 360.0) * M_PI_180;
			}
			while (lonlatCoord[0] * M_180_PI > 180.0)
			{
				lonlatCoord[0] = (lonlatCoord[0] * M_180_PI - 360.0) * M_PI_180;
			}
		}

		result[0] = lonlatCoord[0];
		result[1] = lonlatCoord[1];

	}
}

__device__ void getCenter(int quadNum, int i, int j, int level, double* result, double* triParaments_d, double* vertTable_d)
{
	double cent[2] = { 0.0,0.0 };
	invQuantify(i, j, cent);

	//long double r = M_SQRT3 / 3;
	//long double r2 = r / 2;
	double trans[2] = { 0.0,0.0 };
	double rotDeg = 0.0;
	double scaleFac = powf(2, level);

	convertTypeAddress(cent, scaleFac, trans, rotDeg, quadNum, result, triParaments_d, vertTable_d);
}

__device__ void code2Geo(int basePartition, int row, int col, int level, double& lon, double& lat, double* triParaments_d, double* vertTable_d)
{
	double result[2] = { 0.0,0.0 };
	getCenter(basePartition, row, col, level, result, triParaments_d, vertTable_d);

	lon = result[0] * M_180_PI;
	lat = result[1] * M_180_PI;
}

//9DEMȡǰγȵĸֵ߳
__device__  double get_elev_near_min(double lon, double lat, _tile_info* tinfo_d)
{
	int ic = (lon - tinfo_d->topleft[0]) * tinfo_d->cellSizeT;
	int ir = (tinfo_d->topleft[1] - lat) * tinfo_d->cellSizeT;

	ic = ic < 0 ? 0 : ic;
	ir = ir < 0 ? 0 : ir;

	int ix = (ic / tinfo_d->ndim);
	int iy = (ir / tinfo_d->ndim);

	// clip.tile.
	ix = ix > 2 ? 2 : ix;
	iy = iy > 2 ? 2 : iy;

	int ic1 = ic - ix * tinfo_d->ndim;
	int ir1 = ir - iy * tinfo_d->ndim;
	ic1 = ic1 > tinfo_d->ndim - 1 ? tinfo_d->ndim - 1 : ic1;
	ir1 = ir1 > tinfo_d->ndim - 1 ? tinfo_d->ndim - 1 : ir1;

	int dc = ic1 + 1 < tinfo_d->ndim ? 1 : 0;
	int dr = ir1 + 1 < tinfo_d->ndim ? tinfo_d->ndim : 0;

	short* ptile = tinfo_d->demtiles33[ix + iy * 3];
	short* pe0 = ptile + ic1 + ir1 * tinfo_d->ndim;
	short* pe1 = pe0 + dc;
	short* pe2 = pe0 + dr;
	short* pe3 = pe2 + dc;

	short selev = *pe0 < *pe1 ? *pe0 : *pe1;
	selev = selev > *pe2 ? *pe2 : selev;
	selev = selev > *pe3 ? *pe3 : selev;

	return selev * 1.0;

}

__global__ void dem2Hex2(int quad, double* d_lonArray, double* d_latArray, double* d_elev, _tile_info* tinfo_d, double* triParaments_d, double* vertTable_d)
{
	//printf("%d", tinfo_d->ndim);
	for (int i = blockDim.x * blockIdx.x + threadIdx.x; i < 2048 * 2048; i += blockDim.x * gridDim.x)
	{
		double lon = 0, lat = 0;
		code2Geo(quad, d_lonArray[i], d_latArray[i], 18, lon, lat, triParaments_d, vertTable_d);
		double elev = get_elev_near_min(lon, lat, tinfo_d);
		d_elev[i] = elev;
	}
}

//ȡ3*3ؿĸ߳ϢԼؿtransϢγȷΧԪС
//ȡʧ˵ûд˾γµļʾ˾γȫǺ
int load_one_tile(int ilon, int ilat, int ishp, void* buffer, double* ptransform)
{
	char fmts[128];
	sprintf(fmts, "%s/%s%s", dir_read, str_namefmts[ishp], str_postfmt);
	char path1[256];
	sprintf(path1, fmts, ilat, ilon);

	// test.check.file.exist.
	FILE* pfile = fopen(path1, "rb");
	if (pfile == 0)
	{
		//	printf("...err.path %s\n", path1);
		return 0;
	}
	fclose(pfile);
	//
	//printf("....succ.load.one.%s\n", path1);


	GDALDataset* poDataset = (GDALDataset*)GDALOpen(path1, GA_ReadOnly);
	if (poDataset == 0)
	{
		//	printf("...errr.open.%s\n", path1);
		return 0;
	}
	//printf("...succ.open.%s\n", path1);

	int xSize = poDataset->GetRasterXSize();
	int ySize = poDataset->GetRasterYSize();

	GDALRasterBand* poBand = poDataset->GetRasterBand(1);
	int nXSize = poBand->GetXSize();
	int nYSize = poBand->GetYSize();

	// ???
	GDALDataType itype = poBand->GetRasterDataType();
	//	ibps_dat = GDALGetDataTypeSizeBytes(itype);
	//	ipitch_tile = ibps_dat * nXSize;
	//	isize_tile = nXSize;

	CPLErr err = poBand->RasterIO(GF_Read, 0, 0, nXSize, nYSize, buffer, nXSize, nXSize, GDT_Int16, 0, 0);

	poDataset->GetGeoTransform(ptransform);

	const char* cproj = poDataset->GetProjectionRef();
	if (projdesc && cproj)
	{
		strcpy(projdesc, cproj);
		//	printf("...projinfo: %s\n", projdesc);
	}

	// clean.gdal.
	GDALClose(poDataset);


	return 1;
}

//γȹ黯ΪļвûиֵжilonilatǶϱγеһ
int lonlat_ishp(int& ilon, int& ilat)
{
	int ishp = 0;
	int E = 1;
	int N = 1;

	if (ilon >= 0)
	{
		if (ilon == 180)
		{
			//	ilon = 360 - ilon;
			E = 0;
		}
	}
	else if (ilon < -180)
	{
		ilon = 360 + ilon;
	}
	else
	{
		ilon = -ilon;
		E = 0;
	}

	if (ilat < 0)
	{
		ilat = -ilat;
		N = 0;
	}


	if (E == 1)
		ishp = 0;
	else
		ishp = 1;

	if (N == 0)
		ishp += 2;

	return ishp;
}

//ȡ3*3ؿ
int load_tile33_all(int ilon, int ilat)
{
	static double transforms[6 * 9];
	int nloaded = 0;

	// new.tile33.
	for (int i = 0; i < 9; ++i)
	{
		int ir = i / 3;
		int ic = i - ir * 3;
		//ilon1 ilat1Ϊ3*3ؿӦľγ
		int ilon1 = ilon + ic - 1;
		int ilat1 = ilat + 1 - ir;
		int ishp = lonlat_ishp(ilon1, ilat1);
		//iloadedΪ0ʾilon1 ilat1ؿȫΪϱ
		//ȡ߳ݺtrans(ϽǾγȣԪС)
		int iloaded = load_one_tile(ilon1, ilat1, ishp, h_elvtiles33[i], transforms + i * 6);
		if (iloaded == 1)
		{
			nloaded += 1;
			imsktile33[i] = 1;
		}
		else
		{
			//ڵĻ򽫴˵ؿĸֵ߳ȫֵΪ0
			memset(h_elvtiles33[i], 0, map_bsize(short));
			imsktile33[i] = 0;
		}
	}

	//ĿɹĶȡˣǺϱ飩
	if (imsktile33[4] == 1)
	{
		double* ptrans = transforms + 4 * 6;
		tileinfo.cellSize = ptrans[1];
		//	tileinfo.cellSizeT = 1.0 / tileinfo.cellSize;
		tileinfo.cellSizeT = TILE_SIZE;// - 1;
		memcpy(tileinfo.transform, ptrans, sizeof(double) * 6);
	}
	else // 
	{
		double cellsize = 1.0 / TILE_SIZE;
		double trans[6] = { ilon * 1.0, cellsize, 0, ilat + 1.0, 0, -cellsize };
		memcpy(tileinfo.transform, trans, sizeof(double) * 6);
	}


	tileinfo.topleft[0] = ilon - 1;
	tileinfo.topleft[1] = ilat + 2;
	tileinfo.btnright[0] = ilon + 2;
	tileinfo.btnright[1] = ilat - 1;
	tileinfo.loncen = tileinfo.topleft[0] + 1.5;
	tileinfo.latcen = tileinfo.topleft[1] - 1.5;
	tileinfo.ndim = TILE_SIZE;
	tileinfo.lonext = 3;
	tileinfo.latext = 3;

	inew_cols = 3;

	return nloaded;
}

void loadGridData(int baseQuad, int rowIndex, int colIndex, double* lon, double* lat)
{
	unsigned long long maxRow = pow(2, level);
	unsigned long long maxCol = pow(2, level);
	unsigned long long maxFaces = maxRow * maxCol;
	double temLon = 0, temLat = 0;
	long index = 0;

	for (int i = rowIndex * 2048; i < (rowIndex + 1) * 2048; i++)
	{
		for (int j = colIndex * 2048; j < (colIndex + 1) * 2048; j++)
		{
			//code2Geo(baseQuad, i, j, level, temLon, temLat);
			/*unsigned long long id = (baseQuad - 1) * maxFaces + i * maxCol + j + 1;
			DGGridElement* code;
			pCodes->getRCCode(id, Face, 0, code);
			pCodes->getCenter(code, temLon, temLat);*/
			lon[index] = i;
			lat[index] = j;
			index++;
			/*delete code;
			code = NULL;*/
		}
	}
	//printf("load grid data success!\n");
}

void bufferInit()
{
	memset(values, 0, sizeof(values));


	lonArray = (double*)malloc(gridNumSize);
	latArray = (double*)malloc(gridNumSize);
	elev_h = (double*)malloc(gridNumSize);

	for (int i = 0; i < 9; i++)
	{
		h_elvtiles33[i] = malloc(map_bsize(short));

		void* buf = 0;
		cudaMalloc(&buf, map_bsize(short));
		tileinfo.demtiles33[i] = (short*)buf;
	}


	CHECK(cudaMalloc((double**)&vertTable_d, vertTriSize));
	CHECK(cudaMalloc((double**)&triParaments_d, triSize));

	CHECK(cudaMalloc((double**)&lonArray_d, gridNumSize));
	CHECK(cudaMalloc((double**)&latAttay_d, gridNumSize));
	CHECK(cudaMalloc((double**)&elev_d, gridNumSize));
	CHECK(cudaMalloc((void**)&tileinfo_d, sizeof(_tile_info)));
}

void bufferDelete()
{
	sfree(lonArray);
	sfree(latArray);
	sfree(elev_h);
	for (int i = 0; i < 9; i++)
	{
		sfree(h_elvtiles33[i]);
		cufree(tileinfo.demtiles33[i]);
	}
	cufree(lonArray_d);
	cufree(latAttay_d);
	cufree(elev_d);
	cufree(tileinfo_d);
	cufree(vertTable_d);
	cufree(triParaments_d);
	cufree(values);
	cufree(controlPointLon);
	cufree(controlPointLat);
	cufree(controlPointRow);
	cufree(controlPointCol);
}

int map_buffer(void* dst, void* src, int bsize, int itodevice)
{
	cudaError_t status = cudaSuccess;
	if (itodevice == 1)
		status = cudaMemcpy(dst, src, bsize, cudaMemcpyHostToDevice);
	else
		status = cudaMemcpy(dst, src, bsize, cudaMemcpyDeviceToHost);

	if (status != cudaSuccess)
	{
		if (itodevice == 1)
			printf("....errr.map.buffer.to.device...ecode %d\n", status);
		else
			printf("....errr.map.buffer.to.host...ecode %d\n", status);

	}
	else
	{
		//	if(itodevice==1)
		//		printf("....succ.map.buffer.to.device....\n");
		//	else if(itodevice==0)
		//		printf("....succ.map.buffer.to.host....\n");
	}

	return status == 0 ? 1 : 0;

}

int upd_demtile33_all()
{
	for (int i = 0; i < 9; ++i)
		if (map_buffer(tileinfo.demtiles33[i], h_elvtiles33[i], map_bsize(short), 1) == 0)
			return 0;

	//printf("...upd_demtile33(*).done...\n");

	return 1;
}

int cudop_upd_tileinfo(_tile_info* pinfo)
{
	return map_buffer(tileinfo_d, pinfo, sizeof(_tile_info), 1);
}

//кżʮĪ
int calcMorton(int row, int col)
{
	int m = 0;
	for (size_t i = 0; i < 32; i++)
	{
		m |= (col & 1 << i) << i | (row & 1 << i) << (i + 1);
	}
	return m;
}
//ʮĪк
void fromMorton2ij(int m, int& row, int& col)
{
	row = 0;
	col = 0;
	for (int i = 0; i < 32; i++)
	{
		col |= (m & 1 << (i + i)) >> i;
	}
	for (int i = 0; i < 32; i++)
	{
		row |= (m & 1 << (i + i + 1)) >> (i + 1);
	}
}

double lon, lat, lon_min = 1000, lat_min = 1000, lon_max = -1000, lat_max = -1000;
//ջοкűӦļ
int saveDEM2HexFile(int baseP, int row, int col)
{
	int morton = calcMorton(row, col);
	string mortonName = to_string(morton);
	while (mortonName.size() < 5)
	{
		mortonName = "0" + mortonName;
	}
	string fileName = to_string(baseP - 1) + mortonName;
	while (fileName.size() < 13)
	{
		fileName = fileName + "0";
	}
	char path[256];

	sprintf(path, "%s/%s/%s%s", dir_save, to_string(baseP - 1).c_str(), fileName.c_str(), save_format);

	ofstream ofs;
	ofs.open(path, ios::out | ios::binary);
	if (!ofs.is_open())
	{
		printf("open output file failed.\n");
		return 0;
	}

	ofs << "Grid Res:" << level << endl;
	ofs << "Base Partition:" << baseP << endl;
	ofs << "MinRow is:" << row * 2048 << endl;
	ofs << "MaxRow is:" << row * 2048 + 2048 << endl;
	ofs << "MinCol is:" << col * 2048 << endl;
	ofs << "MaxCol is:" << col * 2048 + 2048 << endl;
	ofs << "MinZ is:" << 0 << endl;
	ofs << "MaxZ is:" << 0 << endl;

	
	code2Geo_H(baseP, row * 2048, col * 2048, 18, lon, lat);
	lon_min = lon < lon_min ? lon : lon_min;
	lat_min = lat < lat_min ? lat : lat_min;
	lon_max = lon > lon_max ? lon : lon_max;
	lat_max = lat > lat_max ? lat : lat_max;

	code2Geo_H(baseP, (row+1) * 2048, col * 2048, 18, lon, lat);
	lon_min = lon < lon_min ? lon : lon_min;
	lat_min = lat < lat_min ? lat : lat_min;
	lon_max = lon > lon_max ? lon : lon_max;
	lat_max = lat > lat_max ? lat : lat_max;

	code2Geo_H(baseP, (row + 1) * 2048, (col+1) * 2048, 18, lon, lat);
	lon_min = lon < lon_min ? lon : lon_min;
	lat_min = lat < lat_min ? lat : lat_min;
	lon_max = lon > lon_max ? lon : lon_max;
	lat_max = lat > lat_max ? lat : lat_max;

	code2Geo_H(baseP, row  * 2048, (col+1) * 2048, 18, lon, lat);
	lon_min = lon < lon_min ? lon : lon_min;
	lat_min = lat < lat_min ? lat : lat_min;
	lon_max = lon > lon_max ? lon : lon_max;
	lat_max = lat > lat_max ? lat : lat_max;

	std::cout << "max lon:" << lon_max << std::endl;
	std::cout << "max lat:" << lat_max << std::endl;
	std::cout << "min lon:" << lon_min << std::endl;
	std::cout << "min lat:" << lat_min << std::endl;

	for (int i = 0; i < 2048; i++)
	{
		for (int j = 0; j < 2048; j++)
		{
			short data = elev_h[i * 2048 + j];
			ofs.write((char*)&data, sizeof(data));
			//ofs << " ";
		}
		//ofs << endl;
	}
	ofs.close();
	return 1;
}


int main()
{
	CPLSetConfigOption("GDAL_FILENAME_IS_UTF8", "NO");	// ֧·

	GDALAllRegister();  //עе 
	poDriver = GetGDALDriverManager()->GetDriverByName(pszFormat);

	bufferInit(); //CPUGPUڴռ

	clock_t start, end;
	start = clock();


	for (int i = 4; i < 5; i++)
	{
		cout << "StartingCalculatBaseP: " << i << endl;
		double lon, lat;

		for (int m = 90; m < 93; m++)
		{
			for (int n = 85; n < 87; n++)
			{

				loadGridData(i, m, n, lonArray, latArray);

				double lon = 0, lat = 0;
				code2Geo_H(i, m * 2048, n * 2048, level, lon, lat);
				cout << lon << "," << lat << endl;

				int num = load_tile33_all((int)lon, (int)lat);
				if (num == 0)
				{
					continue;
				}

				//cout << num << endl;
				if (upd_demtile33_all() == 0)
					return 0;
				if (cudop_upd_tileinfo(&tileinfo) == 0)  //ؿϢݵGPU
					return 0;
				CHECK(cudaMemcpy(lonArray_d, lonArray, gridNumSize, cudaMemcpyHostToDevice));
				CHECK(cudaMemcpy(latAttay_d, latArray, gridNumSize, cudaMemcpyHostToDevice));
				CHECK(cudaMemcpy(vertTable_d, vertTable, vertTriSize, cudaMemcpyHostToDevice));
				CHECK(cudaMemcpy(triParaments_d, triParaments, triSize, cudaMemcpyHostToDevice));

				dem2Hex2 << <128, 128 >> > (i, lonArray_d, latAttay_d, elev_d, tileinfo_d, triParaments_d, vertTable_d);
				cudaError_t err = cudaGetLastError();
				if (err != cudaSuccess) {
					printf("CUDA Error: %s\n", cudaGetErrorString(err));
				}
				cudaMemcpy(elev_h, elev_d, gridNumSize, cudaMemcpyDeviceToHost);

				saveDEM2HexFile(i, m, n);
			}
		}
		cout << "FinishedCalculatBaseP: " << i << endl;
	}
	end = clock();
	cout << "time: " << (end - start) / CLOCKS_PER_SEC << endl;

	bufferDelete();


	return 0;

}
